//------------------------------------------------------------------------------
//  Qt Rubik's Cube
//  ===============
//
//  A Program by :- David Crofts & 
//                  Harry Balzonya
//
//                  January 10, 2012
//
//                  http://members.iimetro.com.au/~crofts/david.htm
//
//                  dasc1961@netscape.net
//
//------------------------------------------------------------------------------

#include "cube.h"

#include <stdlib.h>

#include <qapplication.h>
#include <qgl.h>
#include <math.h>
#include <qlayout.h>
#include <qpushbutton.h>
#include <qvector.h>
#include <qlabel.h>
#include <qpopupmenu.h>
#include <qmessagebox.h>
#include <qwidgetstack.h>
#include <qdatetime.h>
#include <qfile.h>
#include <qdir.h>
#include <qpainter.h>


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

#if defined(_CC_MSVC_)
#pragma warning(disable:4305)			//truncation from const double to float
#endif


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

const QString profile_dir( ".cube/" );	// directory for storage of profile data

static int n				=    3;		// global that hold current cube size.
static int interval			=   35;  	// refresh GL Widget wait. (milliseconds)

static int  commandButton	=    0;

static bool commandLoad		= false;
static bool commandSave		= false;
static bool commandReset	= false;
static bool commandScramble	= false;

static GLfloat view_rotx	=  0.0;
static GLfloat view_roty	=  0.0;
static GLfloat view_rotz	=  0.0;

static GLfloat view_r		= 45.0;

static GLfloat angle		=  0.0;
static GLfloat angle_inc	=  2 * 3.1415926 / 180;


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

list::list()
{

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void list::empty()
{
	list_node *p;
	list_node *q;

	p = tail;
	while (p != NULL) {
		q = p->prev;
		delete p;
		p = q;
	}
	tail = NULL;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void list::add(int command)
{
	list_node *p;

	p = new list_node;
	p->data = command;
	p->prev = tail;
	tail = p;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

cubie::cubie()
{

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cubie::draw()
{
	int i;
	list_node *p;

	for (i=0; i<6; ++i) {
		if (gl_face[i] != 0) {
			glPushMatrix();
			p = rotations[i].tail;
			while ( p != NULL) {
				switch ( p->data ) {
				case x_90:
					glRotatef(  90.0, 1.0, 0.0, 0.0 );
					break;
				case y_90:
					glRotatef(  90.0, 0.0, 1.0, 0.0 );
					break;
				case z_90:
					glRotatef(  90.0, 0.0, 0.0, 1.0 );
					break;
				case x_minus_90:
					glRotatef( -90.0, 1.0, 0.0, 0.0 );
					break;
				case y_minus_90:
					glRotatef( -90.0, 0.0, 1.0, 0.0 );
					break;
				case z_minus_90:
					glRotatef( -90.0, 0.0, 0.0, 1.0 );
					break;
				}
				p = p->prev;
			}
			glCallList(gl_face[i]);
			glPopMatrix();
		}
	}
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cubie::copy(cubie *c)
{
	int i;

	for (i=0; i<6; ++i) {
		gl_face[i] = c->gl_face[i];
		rotations[i].tail = c->rotations[i].tail;
	}

	initial_x = c->initial_x;
	initial_y = c->initial_y;
	initial_z = c->initial_z;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

cube::cube()
{
	int x, y, z, i;

	for (x=0; x<SIZE; ++x) {
		for (y=0; y<SIZE; ++y) {
			for (z=0; z<SIZE; ++z) {
				for (i=0; i<6; ++i) {
					cubies[x][y][z].gl_face[i] = 0;
					cubies[x][y][z].rotations[i].tail = NULL;
				}
			}
		}
	}
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

cube::~cube()
{
	int i, x, y, z;

	for (x=0; x<SIZE; ++x) {
		for (y=0; y<SIZE; ++y) {
			for (z=0; z<SIZE; ++z) {
				for (i=0; i<6; ++i) {
					cubies[x][y][z].rotations[i].empty();
				}
			}
		}
	}
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::init()
{
	int x, y, z, i;

	static GLfloat agreen  [4] = {0.0, 1.0, 0.0, 1.0 };
	static GLfloat ayellow [4] = {1.0, 1.0, 0.0, 1.0 };

	static GLfloat ared    [4] = {1.0, 0.0, 0.0, 1.0 };
	static GLfloat aorange [4] = {1.0, 0.5, 0.0, 1.0 };

	static GLfloat ablue   [4] = {0.0, 0.0, 1.0, 1.0 };
	static GLfloat awhite  [4] = {1.0, 1.0, 1.0, 1.0 };

#ifndef CUBE_DISABLE_DRAMATIC_LIGHTING

	// improved depth cueing.
    
	static GLfloat position[4] = {0.2, 0.2, -1.0, 1.0 };	// positional.

	glLightf( GL_LIGHT0, GL_QUADRATIC_ATTENUATION, 0.03 );
	glLightModeli( GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE );
	glShadeModel( GL_SMOOTH );

#else

	// approximate original.
	// since there was no distance attenuation and the light was fairly
	// far from the object, a directional light gives a similar result.
    
	static GLfloat position[4] = {0.2, 0.2, 1.0, 0.0 };		// directional.

	glShadeModel( GL_FLAT );

#endif

	glLightfv( GL_LIGHT0, GL_POSITION, position );

	glEnable ( GL_LIGHTING );
	glEnable ( GL_LIGHT0 );

	glEnable ( GL_CULL_FACE );
	glDisable( GL_DEPTH_TEST );		// if cull enabled, no overlap in cube.

	GLfloat lf, lf2, dlf, xf, yf;

	lf = 2.0 / (GLfloat)n;
	dlf = lf * 0.1;
	lf2 = lf - dlf;

	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				for (i=0; i<6; ++i) {
					cubies[x][y][z].gl_face[i] = 0;
					cubies[x][y][z].rotations[i].tail = NULL;
				}
			}
		}
	}
				
	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				
				cubies[x][y][z].initial_x = x;
				cubies[x][y][z].initial_y = y;
				cubies[x][y][z].initial_z = z;

				if (x==0) {			
			
					cubies[x][y][z].gl_face[left] =  glGenLists(1);
					glNewList(cubies[x][y][z].gl_face[left], GL_COMPILE);

					cubies[x][y][z].rotations[left].add(y_minus_90);
					cubies[x][y][z].rotations[left].add(x_minus_90);

					glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, aorange);
					glNormal3f( 0.0, 0.0, 1.0 );
					glBegin( GL_QUADS );

					xf  = (GLfloat)y * lf - 1.0;
					yf  = 1.0 - (GLfloat)z * lf;
															
					glVertex3f( xf + dlf, yf - lf2, 1.0 );
					glVertex3f( xf + lf2, yf - lf2, 1.0 );
					glVertex3f( xf + lf2, yf - dlf, 1.0 );
					glVertex3f( xf + dlf, yf - dlf, 1.0 );	

					glEnd();
					glEndList();
				
				} 
				
				if (x==n-1) {
			
					cubies[x][y][z].gl_face[right] =  glGenLists(1);
					glNewList(cubies[x][y][z].gl_face[right], GL_COMPILE);
					
					cubies[x][y][z].rotations[right].add(y_90);
					cubies[x][y][z].rotations[right].add(x_90);

					glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ared);
					glNormal3f( 0.0, 0.0, 1.0 );
					glBegin( GL_QUADS );

					xf  = (GLfloat)y * lf - 1.0;
					yf  = (GLfloat)z * lf - 1.0;
															
					glVertex3f( xf + dlf, yf + dlf, 1.0 );	
					glVertex3f( xf + lf2, yf + dlf, 1.0 );
					glVertex3f( xf + lf2, yf + lf2, 1.0 );
					glVertex3f( xf + dlf, yf + lf2, 1.0 );

					glEnd();
					glEndList();
				
				} 
				
				if (y==0) {
			
					cubies[x][y][z].gl_face[bottom] =  glGenLists(1);
					glNewList(cubies[x][y][z].gl_face[bottom], GL_COMPILE);
					
					cubies[x][y][z].rotations[bottom].add(x_90);
				
					glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, awhite);
					glNormal3f( 0.0, 0.0, 1.0 );
					glBegin( GL_QUADS );

					xf  = (GLfloat)x * lf - 1.0;
					yf  = (GLfloat)z * lf - 1.0;
															
					glVertex3f( xf + dlf, yf + dlf, 1.0 );	
					glVertex3f( xf + lf2, yf + dlf, 1.0 );
					glVertex3f( xf + lf2, yf + lf2, 1.0 );
					glVertex3f( xf + dlf, yf + lf2, 1.0 );

					glEnd();
					glEndList();

				} 
				
				if (y==n-1) {
			
					cubies[x][y][z].gl_face[top] =  glGenLists(1);
					glNewList(cubies[x][y][z].gl_face[top], GL_COMPILE);
					
					cubies[x][y][z].rotations[top].add(x_minus_90);

					glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ablue);
					glNormal3f( 0.0, 0.0, 1.0 );
					glBegin( GL_QUADS );

					xf  = (GLfloat)x * lf - 1.0;
					yf  = 1.0 - (GLfloat)z * lf;
															
					glVertex3f( xf + dlf, yf - lf2, 1.0 );
					glVertex3f( xf + lf2, yf - lf2, 1.0 );
					glVertex3f( xf + lf2, yf - dlf, 1.0 );
					glVertex3f( xf + dlf, yf - dlf, 1.0 );	

					glEnd();
					glEndList();
				
				} 
				
				if (z==0) {
			
					cubies[x][y][z].gl_face[back] =  glGenLists(1);
					glNewList(cubies[x][y][z].gl_face[back], GL_COMPILE);
					
					cubies[x][y][z].rotations[back].add(x_90);
					cubies[x][y][z].rotations[back].add(x_90);
		
					glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, agreen);
					glNormal3f( 0.0, 0.0, 1.0 );
					glBegin( GL_QUADS );

					xf  = (GLfloat)x * lf - 1.0;
					yf  = 1.0 - (GLfloat)y * lf;
															
					glVertex3f( xf + dlf, yf - lf2, 1.0 );
					glVertex3f( xf + lf2, yf - lf2, 1.0 );
					glVertex3f( xf + lf2, yf - dlf, 1.0 );
					glVertex3f( xf + dlf, yf - dlf, 1.0 );	

					glEnd();
					glEndList();
				
				}
				
				if (z==n-1) {
			
					cubies[x][y][z].gl_face[front] =  glGenLists(1);
					glNewList(cubies[x][y][z].gl_face[front], GL_COMPILE);
					
					glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, ayellow);
					glNormal3f( 0.0, 0.0, 1.0 );
					glBegin( GL_QUADS );

					xf  = (GLfloat)x*lf - 1.0;
					yf  = (GLfloat)y*lf - 1.0;
															
					glVertex3f( xf + dlf, yf + dlf, 1.0 );	
					glVertex3f( xf + lf2, yf + dlf, 1.0 );
					glVertex3f( xf + lf2, yf + lf2, 1.0 );
					glVertex3f( xf + dlf, yf + lf2, 1.0 );

					glEnd();
					glEndList();
				
				}
			}
		}
	}
					
	QString file_name;
	QString num;

    file_name = profile_dir + "cube "+ num.setNum(n) + ".ini";
	
	QFile f( file_name );
	
	if ( f.exists() )
		load();

	glEnable( GL_NORMALIZE );
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::reset()
{
	int i, x, y, z;
	cubie *r;
	cubie *s;
    
	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				for (i=0; i<6; ++i) {
					cubies[x][y][z].rotations[i].empty();
				}
			}
		}
	}

	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				r = &cubies[x][y][z];
				s = &cubies_temp[r->initial_x][r->initial_y][r->initial_z];
				s->copy(r);
			}
		}
	}

	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				cubies[x][y][z].copy(&cubies_temp[x][y][z]);
				if (x==0) {			
					cubies[x][y][z].rotations[left  ].add(y_minus_90);
					cubies[x][y][z].rotations[left  ].add(x_minus_90);
				} 
				if (x==n-1) {
					cubies[x][y][z].rotations[right ].add(y_90);
					cubies[x][y][z].rotations[right ].add(x_90);
				} 
				if (y==0) {
					cubies[x][y][z].rotations[bottom].add(x_90);
				} 
				if (y==n-1) {
					cubies[x][y][z].rotations[top   ].add(x_minus_90);
				} 
				if (z==0) {
					cubies[x][y][z].rotations[back  ].add(x_90);
					cubies[x][y][z].rotations[back  ].add(x_90);
				}
			}
		}
	}
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::save()
{
	int x, y, z, i, j;

	cubie *p;
	list_node *q;

    QString file_name;
    QString num;

	char buf[16];

    file_name = profile_dir + "cube "+ num.setNum(n) + ".ini";
	QFile f( file_name );
	f.open( IO_WriteOnly );
	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				
				p = &cubies[x][y][z];

				sprintf( buf, "%c%c%c\n", 
					'0' + p->initial_x, 
					'0' + p->initial_y, 
					'0' + p->initial_z );
				
				for (i = 0; buf[i] != '\0'; ++i)
					f.putch ( buf[i] );

				for (i=0; i<6; ++i) {

					sprintf( buf, "%d ", p->gl_face[i] );
					
					for (j = 0; buf[j] != '\0'; ++j)
						f.putch ( buf[j] );
				}

				f.putch('\n');

				for (i=0; i<6; ++i) {
					q =  p->rotations[i].tail;
					while (q != NULL) {
						f.putch( '0' + q->data );
						q = q->prev;
					}
	
					f.putch('\n');
				}
				
				f.putch('\n');
			}
		}
	}
    
	f.close(); 
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::load()
{
	int x, y, z, i, j;
	unsigned int len;
	unsigned char cx, cy, cz;

	QString file_name;
    QString num;

	cubie *p;
	
    file_name = profile_dir + "cube "+ num.setNum(n) + ".ini";

	QFile f( file_name );
	
	if ( !f.exists() )
		return;
	
	f.open( IO_ReadOnly );

	len = f.size();
	len /= n*n*n;
	len -= 3+2*n;				// average length of rotations list.

	len = (len+8)*n;			// long enough when rotations all on one line.
								// with a bit extra for good measure.
	char *buf = new char[len];

	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
	
				p = &cubies[x][y][z];
				
				f.readLine( buf, len );

				sscanf( buf, "%c%c%c\n", &cx, &cy, &cz );
					
				p->initial_x = cx - '0';
				p->initial_y = cy - '0';
				p->initial_z = cz - '0';
				
				f.readLine( buf, len );
				
				sscanf( buf, "%d %d %d %d %d %d ", 
					&p->gl_face[0],
					&p->gl_face[1],
					&p->gl_face[2],
					&p->gl_face[3],
					&p->gl_face[4],
					&p->gl_face[5]);
					
				for (j=0; j<6; ++j) {
					
					i = f.readLine( buf, len );
					
					p->rotations[j].empty();
					for ( i = i-2; i >= 0; --i )
						p->rotations[j].add( buf[i] - '0' );

				}

				f.readLine( buf, len );
			}
		}
	}

	f.close();
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::draw()
{
	int x, y, z, i, j;
    
	if (commandSave) {
		save();
		commandSave = false;
	}
	
	if (commandLoad) {
		load();
		commandLoad = false;
	}
	
	if (commandReset) {
		reset();
		commandReset = false;
	}
	
	if (commandScramble) {
		scramble();
		commandScramble = false;
	}
	
	if (commandButton) {
		commandButton--;
		i = commandButton % n;
		j = commandButton / n;
		switch (j) {
		case 0:
			screw_x(i, x_minus_90);
			break;
		case 1:
			screw_x(i, x_90);
			break;
		case 2:
			screw_y(i, y_minus_90);
			break;
		case 3:
			screw_y(i, y_90);
			break;
		case 4:
			screw_z(i, z_minus_90);
			break;
		case 5:
			screw_z(i, z_90);
			break;
		}
		commandButton = 0;
	}

    glPushMatrix();

	view_rotx = view_r*sin(angle);
	view_roty = view_r*cos(angle);
	view_rotz = 0.0;

    glRotatef( view_rotx, 1.0, 0.0, 0.0 );
    glRotatef( view_roty, 0.0, 1.0, 0.0 );
    glRotatef( view_rotz, 0.0, 0.0, 1.0 );

	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				cubies[x][y][z].draw();
			}
		}
	}
 
    glPopMatrix();
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::screw_x(int x_plane, int command)
{
	int i, y, z;
	
	for (y=0; y<n; ++y) {
		for (z=0; z<n; ++z) {
			for (i=0; i<6; ++i) {
				cubies[x_plane][y][z].rotations[i].add(command);
			}
		}
	}

	for (y=0; y<n; ++y) {
		for (z=0; z<n; ++z) { 
			cubies_slice[y][z].copy(&cubies[x_plane][y][z]);
		}
	}

	if (command == x_90) {
		
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				cubies[x_plane][y][z].copy(&cubies_slice[z][n-y-1]);
			}
		}
	
	} else if (command == x_minus_90) {
		
		for (y=0; y<n; ++y) {
			for (z=0; z<n; ++z) {
				cubies[x_plane][y][z].copy(&cubies_slice[n-z-1][y]);
			}
		}

	}
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::screw_y(int y_plane, int command)
{
	int i, x, z;
	
	for (x=0; x<n; ++x) {
		for (z=0; z<n; ++z) {
			for (i=0; i<6; ++i) {
				cubies[x][y_plane][z].rotations[i].add(command);
			}
		}
	}

	for (x=0; x<n; ++x) {
		for (z=0; z<n; ++z) {
			cubies_slice[x][z].copy(&cubies[x][y_plane][z]);
		}
	}

	if (command == y_90) {

		for (x=0; x<n; ++x) {
			for (z=0; z<n; ++z) {
				cubies[x][y_plane][z].copy(&cubies_slice[n-z-1][x]);
			}
		}

	} else if (command == y_minus_90) {
		
		for (x=0; x<n; ++x) {
			for (z=0; z<n; ++z) {
				cubies[x][y_plane][z].copy(&cubies_slice[z][n-x-1]);
			}
		}

	}
}

//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::screw_z(int z_plane, int command)
{
	int i, x, y;
	
	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			for (i=0; i<6; ++i) {
				cubies[x][y][z_plane].rotations[i].add(command);
			}
		}
	}

	for (x=0; x<n; ++x) {
		for (y=0; y<n; ++y) {
			cubies_slice[x][y].copy(&cubies[x][y][z_plane]);
		}
	}

	if (command == z_90) {

		for (x=0; x<n; ++x) {
			for (y=0; y<n; ++y) {
				cubies[x][y][z_plane].copy(&cubies_slice[y][n-x-1]);
			}
		}

	} else if (command == z_minus_90) {
		
		for (x=0; x<n; ++x) {
			for (y=0; y<n; ++y) {
				cubies[x][y][z_plane].copy(&cubies_slice[n-y-1][x]);
			}
		}

	}
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void cube::scramble()
{
	int i, rand_axis, rand_command, rand_plane;

	reset();

	for (i=0; i<16*(n-1); ++i) {

		rand_axis = rand()%3;
		rand_command = rand()%3;
		rand_plane = rand()%n;

		switch (rand_axis) {
		case 0:
			switch (rand_command) {
			case 0:
				screw_x(rand_plane, x_90);
				break;
			case 1:
				screw_x(rand_plane, x_90);
				screw_x(rand_plane, x_90);
				break;
			case 2:
				screw_x(rand_plane, x_minus_90);
				break;
			}
			break;
		case 1:
			switch (rand_command) {
			case 0:
				screw_y(rand_plane, y_90);
				break;
			case 1:
				screw_y(rand_plane, y_90);
				screw_y(rand_plane, y_90);
				break;
			case 2:
				screw_y(rand_plane, y_minus_90);
				break;
			}
			break;
		case 2:
			switch (rand_command) {
			case 0:
				screw_z(rand_plane, z_90);
				break;
			case 1:
				screw_z(rand_plane, z_90);
				screw_z(rand_plane, z_90);
				break;
			case 2:
				screw_z(rand_plane, z_minus_90);
				break;
			}
			break;
		}
	}
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeButton::CubeButton( QWidget *parent ) : QPushButton( parent )
{

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeButton::drawButtonLabel( QPainter *p )
{
    int y = (t-1)/n;
	
	QRect r = rect();
    
	if ( y > 3 )
		p->setPen( QPen( black, 2 ) );
	else
		p->setPen( QPen( white, 2 ) );

	p->drawLine( 
		r.topLeft() + QPoint(  8, 12 ),
		r.topLeft() + QPoint( 16, 12 )
	);

	if ( !(y%2) )
		p->drawLine( 
			r.topLeft() + QPoint( 12,  8 ),
			r.topLeft() + QPoint( 12, 16 ) 
		);

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeGLWidget::CubeGLWidget( QWidget *parent, const char *name )
     : QGLWidget( parent, name )
{
	cubeGL = new cube;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeGLWidget::~CubeGLWidget()
{
	delete	cubeGL;
	killTimer( timer );
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeGLWidget::initializeGL()
{
    cubeGL->init();
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeGLWidget::perspective( GLfloat fovy, GLfloat aspect, GLfloat xnear, GLfloat xfar )
{
	// more intuitive than frustum.
	// this will help make world positions more sensible.
	// assert 0 < fovy < 180
	// assert 0 < xnear < xfar
    
	GLfloat bottom;
    GLfloat top;
    GLfloat right;
    GLfloat left;
	
    bottom = tan( fovy * 0.00872664611 );
	bottom = bottom * xnear;
	top = -bottom;
	right = bottom * aspect;
	left = -right;

	glFrustum( left, right, top, bottom, xnear, xfar );

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeGLWidget::resizeGL( int width, int height )
{

	glViewport( 0, 0, width, height );
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();

    perspective( 45.0, (GLfloat)width/height, 1.0, 10.0 ); // view_r = 45.0
    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();
    glTranslatef( 0.0, 0.0, -5.0 ); // -55.0

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeGLWidget::paintGL()
{    

	GLint viewport[4];
	GLint x, y, w, h;

	glClear( GL_COLOR_BUFFER_BIT );

	angle = fmod( angle + angle_inc, 6.2831852 ); // preserves angle on animation reset.

	cubeGL->draw();					// 1st pass.

#ifndef CUBE_DISABLE_REARVIEW

	glGetIntegerv( GL_VIEWPORT, viewport );

	x = viewport[0];
	y = viewport[1];
	w = viewport[2];
	h = viewport[3];

	glViewport( x, y + (h-1) - h/4, w/4, h/4 ); // upper left (GL yaxis inverted).

	glFogi( GL_FOG_MODE, GL_EXP2 );
	glFogf( GL_FOG_DENSITY, 0.20 );

	glEnable( GL_FOG );				// dim image. reduce aliasing.
	glCullFace( GL_FRONT );			// mirrored faces will back facing.

	glPushMatrix();
	glScalef( 1.0, 1.0, -1.0 );		// mirror.

	cubeGL->draw();					// 2nd pass.

	glPopMatrix();

	glCullFace( GL_BACK );			// restore initial states.
	glDisable( GL_FOG );
	glViewport( x, y, w, h );

#endif

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeGLWidget::timerEvent(QTimerEvent *e)
{
    if (e->timerId() == timer)
		updateGL();
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeGLWidget::kill_timer()
{
    killTimer( timer );
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeGLWidget::start_timer()
{
	timer = startTimer( interval );
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeWidget::CubeWidget( QWidget *parent, const char *name ) : QWidget( parent, name )
{
	QVBoxLayout *l = new QVBoxLayout( this, 0 );

	cubeGLWidget = new CubeGLWidget( this );
    l->addWidget( cubeGLWidget );
    
	QHBoxLayout * h = new QHBoxLayout;
	l->addLayout( h );
    
	QGridLayout * grid1 = new QGridLayout( 3, 1, 0 );
	h->addLayout( grid1 );
	
	leftButton = new QPushButton( "&Left", this );
	leftButton->setMinimumHeight(52);
	leftButton->setMinimumWidth(104);
	grid1->addWidget( leftButton, 0, 0 );
    connect( leftButton, SIGNAL(clicked()), SLOT(leftButtonClicked()) );
	
	upButton = new QPushButton( "&Up", this );
	upButton->setMinimumHeight(52);
	upButton->setMinimumWidth(104);
	grid1->addWidget( upButton, 1, 0 );
    connect( upButton, SIGNAL(clicked()), SLOT(upButtonClicked()) );
	
	downButton = new QPushButton( "&Down", this );
	downButton->setMinimumHeight(52);
	downButton->setMinimumWidth(104);
	grid1->addWidget( downButton, 2, 0 );
    connect( downButton, SIGNAL(clicked()), SLOT(downButtonClicked()) );
	
    buttonGrid = new CubeButtonGrid( this );
	buttonGrid->setMaximumWidth( n*25 + (n+1)*2 );
	h->addWidget( buttonGrid );

	QGridLayout * grid2 = new QGridLayout( 3, 1, 0 );
	h->addLayout( grid2 );
	
	rightButton = new QPushButton( "&Right", this );
	rightButton->setMinimumHeight(52);
	rightButton->setMinimumWidth(104);
	grid2->addWidget( rightButton, 0, 0 );
    connect( rightButton, SIGNAL(clicked()), SLOT(rightButtonClicked()) );
	
	clockwiseButton = new QPushButton( "&Clockwise", this );
	clockwiseButton->setMinimumHeight(52);
	clockwiseButton->setMinimumWidth(104);
	grid2->addWidget( clockwiseButton, 1, 0 );
    connect( clockwiseButton, SIGNAL(clicked()), SLOT(clockwiseButtonClicked()) );
	
	anticlockwiseButton = new QPushButton( "&Anti-Clockwise", this );
	anticlockwiseButton->setMinimumHeight(52);
	anticlockwiseButton->setMinimumWidth(104);
	grid2->addWidget( anticlockwiseButton, 2, 0 );
    connect( anticlockwiseButton, SIGNAL(clicked()), SLOT(anticlockwiseButtonClicked()) );

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeWidget::~CubeWidget()
{

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWidget::downButtonClicked()
{
	for (int i=0; i<n; ++i)
		cubeGLWidget->cubeGL->screw_x(i, x_90);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWidget::upButtonClicked()
{
	for (int i=0; i<n; ++i)
		cubeGLWidget->cubeGL->screw_x(i, x_minus_90);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWidget::rightButtonClicked()
{
	for (int i=0; i<n; ++i)
		cubeGLWidget->cubeGL->screw_y(i, y_90);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWidget::leftButtonClicked()
{
	for (int i=0; i<n; ++i)
		cubeGLWidget->cubeGL->screw_y(i, y_minus_90);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWidget::clockwiseButtonClicked()
{
	for (int i=0; i<n; ++i)
		cubeGLWidget->cubeGL->screw_z(i, z_minus_90);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWidget::anticlockwiseButtonClicked()
{
	for (int i=0; i<n; ++i)
		cubeGLWidget->cubeGL->screw_z(i, z_90);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeWindow::CubeWindow( QWidget* parent, const char* name, WFlags f )
	: QMainWindow( parent, name, f )
{
	int i;
	char buf[32];

	QTime t = QTime::currentTime();							// set random seed
    srand( t.hour()*12+t.minute()*60+t.second()*60 );
    
	if (!QDir(profile_dir).exists() ) 
		QDir().mkdir(profile_dir); 
 
	QFile qf( profile_dir + "cube n.ini" );
	if ( qf.exists() && qf.open( IO_ReadOnly ) ) {
		qf.readLine( buf, 16 );
		sscanf( buf, "%d %f", &n, &angle_inc );
		qf.close();
	}

	QPopupMenu *file = new QPopupMenu( this );
    file->insertItem( "Save Cube",           this, SLOT(save()),     CTRL+Key_S );
    file->insertItem( "Load Cube",           this, SLOT(load()),     CTRL+Key_L );
    file->insertSeparator();
    file->insertItem( "Reset Cube",          this, SLOT(reset()),    CTRL+Key_R );
    file->insertSeparator();
    file->insertItem( "Scramble Cube",       this, SLOT(scramble()), CTRL+Key_M );
    file->insertSeparator();
    file->insertItem( "Go To   2x2x2  Cube", this, SLOT(new2()),     CTRL+Key_2 );
    file->insertItem( "Go To   3x3x3  Cube", this, SLOT(new3()),     CTRL+Key_3 );
    file->insertItem( "Go To   4x4x4  Cube", this, SLOT(new4()),     CTRL+Key_4 );
    file->insertItem( "Go To   5x5x5  Cube", this, SLOT(new5()),     CTRL+Key_5 );
    file->insertItem( "Go To   6x6x6  Cube", this, SLOT(new6()),     CTRL+Key_6 );
    file->insertItem( "Go To   7x7x7  Cube", this, SLOT(new7()),     CTRL+Key_7 );
    file->insertItem( "Go To   8x8x8  Cube", this, SLOT(new8()),     CTRL+Key_8 );
    file->insertSeparator();
    file->insertItem( "E&xit",               qApp, SLOT(quit()),     CTRL+Key_Q );
    
    QPopupMenu *speed = new QPopupMenu( this );
    speed->insertItem( "&Slow",   this, SLOT(slow()) );
    speed->insertItem( "&Medium", this, SLOT(medium()) );
    speed->insertItem( "&Fast",   this, SLOT(fast()) );

    QPopupMenu *view = new QPopupMenu( this );
    view->insertItem( "&Wobble Speed", speed, CTRL+Key_S );

    QPopupMenu *help = new QPopupMenu( this );
    help->insertItem( "&About",    this, SLOT(about()), CTRL+Key_H );
    help->insertItem( "About &Qt", this, SLOT(aboutQt()) );

	menuBar()->insertItem( "&File", file );
	menuBar()->insertItem( "&View", view );
	menuBar()->insertItem( "&Help", help );

	CubeWidgetStack = new QWidgetStack( this );
	setCentralWidget( CubeWidgetStack );
	
	for (i=2; i<=SIZE; ++i)
		CubeWidgets[i] = NULL;

	CubeWidget *cw = new CubeWidget( this );
	CubeWidgets[n] = cw;
	cw->setMinimumWidth(430);
	cw->setMinimumHeight(460);

	CubeWidgetStack->addWidget( cw, n );
	CubeWidgetStack->raiseWidget( n );
	CubeWidgets[n]->cubeGLWidget->start_timer();
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeWindow::~CubeWindow()
{
	char buf[32];
	int i;
	
	QFile f( profile_dir + "cube n.ini" );
	f.open( IO_WriteOnly );
	sprintf( buf, "%d %f", n, angle_inc );
	for ( i=0; buf[i] != '\0'; ++i)
		f.putch( buf[i] );
    f.close();
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::load()
{
	commandLoad = true;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::save()
{
	commandSave = true;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::newCentralWidget(int size)
{
	CubeWidgets[n]->cubeGLWidget->kill_timer();

	n = size;
	
	if (CubeWidgets[n] == NULL) {

		CubeWidget *cw = new CubeWidget( this );
		CubeWidgets[n] = cw;
		cw->setMinimumWidth(430);
		cw->setMinimumHeight(460);

		CubeWidgetStack->addWidget( cw, n );
	}

	CubeWidgetStack->raiseWidget( n );
	CubeWidgets[n]->cubeGLWidget->start_timer();

}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::reset()
{
	commandReset = true;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::scramble()
{
	commandScramble = true;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::new2()
{
	newCentralWidget(2);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::new3()
{
	newCentralWidget(3);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::new4()
{
	newCentralWidget(4);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::new5()
{
	newCentralWidget(5);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::new6()
{
	newCentralWidget(6);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::new7()
{
	newCentralWidget(7);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::new8()
{
	newCentralWidget(8);
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::slow()
{
	angle_inc = 2 * 3.1415926 / 720;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::medium()
{
	angle_inc = 2 * 3.1415926 / 180;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::fast()
{
	angle_inc = 2 * 3.1415926 / 90;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::about()
{
    QMessageBox::about( this, 
		"Qt Rubik's Cube",
		"\n\n"
		"         (C)  David Crofts & Harry Balzonya                    \n\n"
		"                January 10, 2012\n\n"
		"                23 Brisbane Street\n"
		"                Berwick Victoria 3806\n"
		"                Australia\n\n"
		"                dasc1961@netscape.net\n" );
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeWindow::aboutQt()
{
    QMessageBox::aboutQt( this, "Qt Rubik's Cube" );
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeButtonGrid::CubeButtonGrid( QWidget *parent, const char *name )
    : QWidget( parent, name )
{
    int x, y, i;
	
	buttons = new CubeButtons( n*6 );
    QGridLayout * grid = new QGridLayout( this, 6, n, 2 );

	QPalette red   ( QColor( 255,   0,   0 ) );
	QPalette blue  ( QColor(   0,   0, 255 ) );
	QPalette yellow( QColor( 255, 255,   0 ) );
		
    for ( y=0; y<6; y++ ) {
	    for ( x=0; x<n; x++ ) {
			CubeButton *cb = new CubeButton( this );

			connect( cb, SIGNAL(clicked()), SLOT(buttonClicked()) );

			if (y<2)  
				cb->setPalette( red );
			else if (y<4)
				cb->setPalette( blue );
			else
				cb->setPalette( yellow );

			cb->setMaximumHeight(25);
			cb->setMaximumWidth(25);
			cb->setMinimumHeight(25);
			cb->setMinimumWidth(25);
			grid->addWidget( cb, y, x );
			i = y*n+x;
			buttons->insert( i, cb );
			buttons->at(i)->setType(i+1);
		}
    }
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

CubeButtonGrid::~CubeButtonGrid()
{
    delete buttons;
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

void CubeButtonGrid::buttonClicked()
{
    int i;
	
	i = buttons->findRef( (CubeButton*)sender() );
	commandButton = buttons->at(i)->type();
}


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

