//------------------------------------------------------------------------------
//  Qt Rubik's Cube
//  ===============
//
//  A Program by :- David Crofts & 
//                  Harry Balzonya
//
//                  January 10, 2012
//
//                  http://members.iimetro.com.au/~crofts/david.htm
//
//                  dasc1961@netscape.net
//
//------------------------------------------------------------------------------

#include <qgl.h>
#include <qpushbutton.h>
#include <qvector.h>
#include <qlcdnumber.h>
#include <qlabel.h>
#include <qmainwindow.h>
#include <qmenubar.h>
#include <qwidgetstack.h>


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

#define	   SIZE				8				// maximum number of cubies per side.

//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

enum { front, back, top, bottom, right, left };

enum { x_90, x_minus_90, y_90, y_minus_90, z_90, z_minus_90 };


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class list_node {
public:
	list_node *prev;
	int data;
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class list {
public:
	list();

	void add(int data);
	void empty();

	list_node *tail;
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class cubie {
public:
	cubie();

	void draw();
	void copy(cubie *c);

	GLint gl_face[6];
	list  rotations[6];
	
	int   initial_x;
	int	  initial_y;
	int	  initial_z;
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class cube {
public:
	cube();
	~cube();

	void init();
	void reset();
	void scramble();
	void draw();
	void load();
	void save();
	void screw_x(int x_plane, int command);
	void screw_y(int y_plane, int command);
	void screw_z(int y_plane, int command);

	cubie cubies[SIZE][SIZE][SIZE];

private:
	cubie cubies_temp[SIZE][SIZE][SIZE];
	cubie cubies_slice[SIZE][SIZE];
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class CubeGLWidget : public QGLWidget
{
    Q_OBJECT
public:
    CubeGLWidget( QWidget *parent=0, const char *name=0 );
    ~CubeGLWidget();

	cube *cubeGL;
	int  timer;
	void start_timer();
	void kill_timer();
    void perspective( GLfloat, GLfloat, GLfloat, GLfloat );

protected:
    void initializeGL();
    void resizeGL( int, int );
    void paintGL();
    void timerEvent( QTimerEvent * );
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class CubeButton : public QPushButton
{
    Q_OBJECT
public:
    CubeButton( QWidget *parent );
    
    int type() const		
    { return t; }
    
    void setType( int type )	
    { t = type; }
    
	void drawButtonLabel (QPainter *);

private:
    int t;
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

typedef QVector<CubeButton>	CubeButtons;


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class CubeButtonGrid : public QWidget
{
    Q_OBJECT
public:
    CubeButtonGrid( QWidget *parent=0, const char *name=0 );
    ~CubeButtonGrid();

private slots:
    void	buttonClicked();

private:
    CubeButtons *buttons;
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class CubeWidget : public QWidget
{
    Q_OBJECT
public:
    CubeWidget( QWidget *parent=0, const char *name=0 );
    ~CubeWidget();

	CubeGLWidget   *cubeGLWidget;

private slots:
    void	leftButtonClicked();
    void	rightButtonClicked();
    void	upButtonClicked();
    void	downButtonClicked();
    void	clockwiseButtonClicked();
    void	anticlockwiseButtonClicked();

private:
    CubeButtonGrid *buttonGrid;
	QPushButton    *leftButton;
    QPushButton    *rightButton;
    QPushButton    *upButton;
    QPushButton    *downButton;
    QPushButton    *clockwiseButton;
    QPushButton    *anticlockwiseButton;
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

class CubeWindow : public QMainWindow
{
	Q_OBJECT
public:
	CubeWindow( QWidget *parent = 0, const char *name = 0, WFlags f = WType_TopLevel );
	~CubeWindow();

	void newCentralWidget(int size);
		
	QWidgetStack *CubeWidgetStack;
	CubeWidget	 *CubeWidgets[SIZE+1];

public slots:
	void load();
	void save();
	void reset();
	void scramble();
	void new2();
	void new3();
	void new4();
	void new5();
	void new6();
	void new7();
	void new8();
	void slow();
	void medium();
	void fast();
	void about();
	void aboutQt();
};


//------------------------------------------------------------------------------
//
//------------------------------------------------------------------------------

